#include <stdio.h>
#include <ldap.h>

int main(void)
{
	LDAP            *ld;
	LDAPMessage     *res, *e;
	int             i, rc;
	char            *a, *dn;
	BerElement      *ptr;
	char            **vals;

	/* open an LDAP session */
	if ( (ld = ldap_init( "ldap.eparana.parana", 389 )) == NULL )
	{
		fprintf( stderr, "erro ao se conectar:");
		return 1;
	}

	/* authenticate as nobody */
	if (( rc = ldap_simple_bind_s( ld, NULL, NULL )) != LDAP_SUCCESS )
	{
		fprintf( stderr, "ldap_simple_bind_s: %s\n", ldap_err2string( rc ));
		ldap_unbind( ld );
		return 1;
	}

	/* search for entries with cn of "Babs Jensen", return all attrs  */
	if (( rc = ldap_search_s( ld, "dc=pr, dc=gov, dc=br", LDAP_SCOPE_SUBTREE, "(uid=alexandrecorreia)", NULL, 0, &res )) != LDAP_SUCCESS )
	{
		fprintf( stderr, "ldap_search_s: %s\n", ldap_err2string( rc ));
		if ( res == NULL )
		{
			ldap_unbind( ld );
			return 1;
		}
	}

	/* step through each entry returned */
	for ( e = ldap_first_entry( ld, res ); e != NULL; e = ldap_next_entry( ld, e ) )
	{
		/* print its name */
		dn = ldap_get_dn( ld, e );
		printf( "uid: %s\n", dn );
		ldap_memfree( dn );

		/* print each attribute */
		for ( a = ldap_first_attribute( ld, e, &ptr ); a != NULL; a = ldap_next_attribute( ld, e, ptr ) )
		{
			printf( "\tattribute: %s\n", a );

			/* print each value */
			vals = ldap_get_values( ld, e, a );

			for ( i = 0; vals[i] != NULL; i++ )
				printf( "\t\tvalue: %s\n", vals[i] );

			ldap_value_free( vals );
			ldap_memfree( a );
		}
		if ( ptr != NULL )
			ber_free( ptr, 0 );
	}
	/* free the search results */
	ldap_msgfree( res );

	/* close and free connection resources */
	ldap_unbind( ld );

	return 0;
}
